;;; **********************************************************************
;;; Copyright (c) 1995 Heinrich Taube and Tobias Kunze.  All rights
;;; reserved. Use and copying of this software and preparation of
;;; derivative works based upon this software are permitted and may be
;;; copied as long as no fees or compensation are charged for use, 
;;; copying, or accessing this software and all copies of this software
;;; include this copyright notice.
;;; **********************************************************************

(in-package :stella)

;;;
;;; table displays each entry with a label and edit buffer. 
;;; entries are lists: (id labelview bufview mode)
;;; mode can be t, :hidden or :readonly
;;;

(defclass table (ccl:view)
  ((entries :initform nil :initarg :entries :accessor table-entries)
   (label-args :initform (list 'table-label)
               :initarg :label-args)
   (value-args :initform (list 'textbuf)
               :initarg :value-args)
   (extent :accessor table-extent :initform (xy 0 0))
   (claims :initarg :claims :accessor table-claims :initform nil)
   (editx :initform nil :initarg :edit-x)  ; x pos for buffers
   (color :accessor table-color :initarg :table-color :initform ccl:*white-color*)) 
  (:default-initargs 
    :view-font *capella-font*
    :view-nick-name ':table))

;(defun set-table-extent (table size)
;  (setf (table-extent table) size)
;  (ccl:set-view-size table size))

(defun set-table-claims (table size)
  (setf (table-claims table) size)
  (ccl:set-view-size table size))

(defconstant +entry-moveable+ 1)
(defconstant +entry-hideable+ 2)
(defconstant +entry-readonly+ 4)
(defconstant +entry-hidden+ 8)
(defconstant +entry-normal-mode+
  (logior +entry-moveable+ +entry-hideable+))

(defclass table-entry (ccl:view)
  ((id :initarg :id :initform nil :accessor table-entry-id)
   (label :initarg :label :initform nil :accessor table-entry-label)
   (editor :initarg :editor :initform nil :accessor table-entry-editor)
   (mode :initarg :mode :initform +entry-normal-mode+
         :accessor table-entry-mode)
   (y :initarg :y :initform 0 :accessor table-entry-y)
   (type :initarg :type :initform nil :accessor table-entry-type)
   (color :accessor table-entry-color :initarg :color :initform nil)
   )
  (:default-initargs :view-font *capella-font*))

(defun make-table-entry (&rest args)
  (apply #'make-instance 'table-entry args))

(defmethod table-entry-table ((view table-entry))
  ;; fix, works only on visible entries!
  (ccl:view-container view))

(defmethod initialize-instance :after ((view table) &rest args)
  (declare (ignore args))
  ;; map over entries, establish labels and buffers as needed.
  (let ((width 0))
    (loop for entry in (slot-value view 'entries)
          do
          (insure-table-entry view entry)
          (setf width
                (max width (+ (view-width (table-entry-label entry))
                              4 5))))
    (unless (slot-value view 'editx)
      (setf (slot-value view 'editx) 
            (xy (+ width 5 4) 0)))
    (let ((claims (table-claims view)))
      (when claims (ccl:set-view-size view claims)))
    (size-table view (xy-x (ccl:view-size view)))))

(defun add-table-entry (table entry &optional (redisplay? t))
  (setf entry (insure-table-entry table entry))
  (setf (table-entries table)
        (nconc (table-entries table) (list entry)))
  (when redisplay?
    (redisplay-changed-table table)
    (ccl:set-current-key-handler
     (ccl:view-window table) (table-entry-editor entry) t)
    ;; not sure if this really belongs here.
    (when (typep table 'ccl::scroller-mixin)
      (let ((scroller (ccl::v-scroller table)))
        (ccl:set-scroll-bar-setting scroller 
                                    (xy-y (ccl:view-position entry)))
        (ccl:dialog-item-action scroller)))))

(defmethod insure-table-entry ((table table) (entry table-entry))
  (let ((label (table-entry-label entry))
        (value (table-entry-editor entry))
        (largs (slot-value table 'label-args))
        (eargs (slot-value table 'value-args))
        class)
    (unless (typep label 'ccl:simple-view)
      (setf class (pop largs))
      (setf label (apply #'make-instance class
                         (if (typep class 'ccl:dialog-item)
                           :dialog-item-text :title)
                         (if (stringp label) label
                             (format nil "~:(~A~)"
                                     (or label (table-entry-id entry))))
                         largs))
      (ccl:set-view-size 
       label (xy (+ 4 (ccl:string-width (item-title label)
                                        (ccl:view-font label)))
                 (font-height (ccl:view-font label))))
      (setf (table-entry-label entry) label))
    (unless (typep value 'ccl:simple-view)
      (setf class (pop eargs))
      (setf (table-entry-editor entry) 
            (apply #'make-instance 
                   (if (table-entry-readonly? entry) 'label class)
                   eargs))
      (form-to-editor (table-entry-editor entry) value)
      (setf value (table-entry-editor entry)))
    (setf (table-entry-y entry)
          (+ 7 (max (table-entry-y entry)
                    (view-height label)
                    (view-height value))))
    (unless (eq entry (ccl:view-container label))
      (ccl:add-subviews entry label))
    (unless (eq entry (ccl:view-container value))
      (ccl:add-subviews entry value))
    (unless (table-entry-color entry)
      (setf (table-entry-color entry) (table-color table)))
    entry))

(defun rename-table-entry (entry newname &optional label)
  (setf (table-entry-id entry) newname)
  (ccl:set-dialog-item-text (table-entry-label entry) 
                            (or label
                                (string-capitalize (string newname)))))

(defmethod size-table ((view table) h &optional y)
  ;; layout table based on width of view and computed height 
  ;; of visible entries. returns new and old sizes
  (let ((entries (slot-value view 'entries))
        (editx (slot-value view 'editx))
        (yfudge 0)         ; dialog border fudge factor
        (size h)
        views hides)
    (setf y (+ yfudge 0))
    (loop for entry in entries
          do 
          (if (table-entry-visible? entry)
            (progn
              (size-table-entry entry (xy 0 y) size editx)
              (incf y (+ yfudge (table-entry-y entry)))
              (push entry views))
            (push entry hides) ))
    (apply #'remove-hidden-entries view hides)
    (apply #'ccl:add-subviews view views)
    (let ((old (ccl:view-size view))
          (new (xy (xy-x h) (if (= y yfudge) 0 y))))
      (setf (table-extent view) new)
      (if (table-claims view)
        (progn
          (ccl::update-scroll-bar-limits view)
          ;(update-scollbar-visibility view)
          )
        (ccl:set-view-size view new))
      (values new old))))

;;;
;;;
;;;

(defmethod form-to-editor ((buffer titled-item-mixin) value)
  (setf (item-title buffer) 
   (if (stringp value) value (if value (prin1-to-string value) "")))
  (ccl:set-view-size 
   buffer (xy 0 (font-height (ccl:view-font buffer)))))

(defmethod form-to-editor ((buffer ccl:dialog-item) value)
  (ccl:set-dialog-item-text 
   buffer 
   (if (stringp value) value (if value (prin1-to-string value) "")))
  (ccl:set-view-size 
   buffer (xy 0 (font-height (ccl:view-font buffer)))))

(defmethod editor-to-form ((entry table-entry) &optional empty=nil?)
  ;; return form and flag: t | nil | :error
  (if (and (table-entry-visible? entry)
           (not (table-entry-readonly? entry)))
    (editor-to-form (table-entry-editor entry) empty=nil?)
    (values nil nil)))

(defmethod editor-to-form ((buffer ccl:dialog-item)
                           &optional empty=nil?)
  ;; return form and flag: t | nil | :error
  (let ((s (textbuf-text buffer)))
    (if (string= s "") 
      (values nil empty=nil?)
      (if (balanced-form? s)
        (values (read-from-string s) t)
        (values s ':error)))))

(defmethod size-table-entry ((view table-entry) pos size editx)
  (let ((buf (table-entry-editor view))
        (label (table-entry-label view)))
    ;; set y position and size of whole entry
    (ccl:set-view-position view pos)
    (ccl:set-view-size view (xy (xy-x size) (table-entry-y view)))
    ;; set label 3 points down to be on the same height as edit view text, too
    (ccl:set-view-position label (xy 0 3))
    ;; set edit view +x in from 0 and -y in from width
    (ccl:set-view-position buf (xy (xy-x editx) 3))
    (ccl:set-view-size buf (xy (- (xy-x size) (xy-x editx) (xy-y editx)
                                  3)
                               (view-height buf)))))

(defun install-view (view new &optional old)
  ;; add new to view after first removing old, if supplied.
  ;; adjust y positions of all subs > new and adjust height 
  ;; of all superior views.
  (let (window)
    (if (null new)
      ;; no new view, just remove old and subtract its height
      ;; from superior views
      (when old
        (setf window (ccl:view-window old))
        (resize-superiors (ccl:view-container old)
                          0 (- (view-height old)))
        (adjust-later-views window
                            (xy-y (ccl:convert-coordinates
                                   (ccl:view-position old)
                                   (ccl:view-container old) 
                                   window))
                            (- (view-height old)))
        (ccl:remove-subviews view old))
      (if (not (eq new old))
        (let ((y 0))
          (setf window (ccl:view-window view))
          (when old
            (setf y (view-height old))
            (adjust-later-views window
                            (xy-y (ccl:convert-coordinates
                                   (ccl:view-position old)
                                   (ccl:view-container old) 
                                   window))
                            (- y))
            (ccl:remove-subviews view old))
          (ccl:set-view-container new view)
           (adjust-later-views window
                              (xy-y (ccl:convert-coordinates
                                     (ccl:view-position new)
                                     view window))
                              (view-height new))
          (resize-superiors view
                            0 (- (view-height new) y)))))))

(defun adjust-later-views (window y dy)
  (ccl:map-subviews
   window
   #'(lambda (v &aux (p (ccl:view-position v)))
       (when (> (xy-y p) y)
         (ccl:set-view-position 
          v (xy (xy-x p) (+ (xy-y p) dy)))))))
   

(defun select-table-editor (table &optional id)
  (if id
    (let ((e (table-find id table :key #'table-entry-id)))
      (if (table-entry-visible? e)
        (ccl:set-current-key-handler (ccl:view-window table)
                                     (table-entry-editor e))
        nil))
    (let ((e (table-find-if #'table-entry-visible? table)))
      (when e 
        (ccl:set-current-key-handler (ccl:view-window table)
                                     (table-entry-editor e))))))

(defun invalidate-table-entry (entry)
  (ccl:invalidate-view (table-entry-label entry))
  (ccl:invalidate-view (table-entry-editor entry)))

(defun hide-table-entry (entry)
  (setf (table-entry-mode entry) 
        (logior (table-entry-mode entry) +entry-hidden+)))

(defun unhide-table-entry (entry)
  (setf (table-entry-mode entry)
        (logandc2 (table-entry-mode entry) +entry-hidden+)))

(defun table-entry-hidden? (entry)
  (logtest (table-entry-mode entry) +entry-hidden+))

(defun table-entry-visible? (entry)
  (not (table-entry-hidden? entry)))

(defun table-entry-readonly? (entry)
  (logtest (table-entry-mode entry) +entry-readonly+))

(defun table-entry-editable? (entry)
  (not (table-entry-readonly? entry)))

(defun map-table (fn view)
  (map nil fn (slot-value view 'entries)))

(defun table-hidden-entries? (view)
  (map-table #'(lambda (p)
                 (when (table-entry-hidden? p)
                   (return-from table-hidden-entries? t)))
             view)
  nil)

(defun table-find (thing table &rest args)
  (apply #'find thing (table-entries table)  args))

(defun table-find-if (fn table &rest args)
  (apply #'find-if fn (table-entries table) args))

(defun table-position (entry table &rest args)
  (apply #'position entry (table-entries table) args))

(defun table-position-if (fn table &rest args)
  (apply #'position-if fn (table-entries table) args))

(defun move-table-entry (table from to)
  (let* ((entries (table-entries table)) ;
         (length (length entries))
         frompos topos)
    (if (numberp from) 
      (setf frompos from from (elt entries from))
      (setf frompos (position from entries)))
    (if (numberp to) 
      (setf topos (min to length) to (elt entries to))
      (setf topos (position to entries)))
    (if (< topos frompos)
      (loop for i from (1- frompos) downto topos 
            do (setf (elt entries (1+ i)) (elt entries i) )
            finally (setf (elt entries topos) from
                          ))
      (if (> topos frompos)
        (loop for i from (1+ frompos) to topos
              do (setf (elt entries (1- i)) (elt entries i))
              finally (setf (elt entries topos) from))))
    (/= topos frompos)))

(defun remove-hidden-entries (view &rest entries)
  (let ((subs (ccl:subviews view)))
    (dolist (e (or entries (slot-value view 'entries)))
      (when (table-entry-hidden? e)
        (when (find e subs)
          (ccl:remove-subviews view e))))))

(defun table-clear (view)
  (map-table
   #'(lambda (e) 
       (ccl:set-dialog-item-text (table-entry-editor e)
                                 ""))
   view))

(defun reshow-table (table &optional allslots?)
  (let ((object (slot-value table 'object))
        redraw)
    (if allslots?
      (if (table-hidden-entries? table)
        (progn
          (map-table #'unhide-table-entry
                     table)
          (setf redraw t))
        nil)
      (if (table-hidden-entries? table)
        nil
        (let ((pars (object-parameters object))) 
          (map-table #'(lambda (p) 
                         (unless (find (table-entry-id p) 
                                       pars :key #'parameter-name)
                           (hide-table-entry p)))
                     table)
          (setf redraw t))))
    (when redraw
      (multiple-value-bind (newsize oldsize)
                           (size-table table (ccl:view-size table))
        (if (or (xy= newsize oldsize) (table-claims table))
          (ccl:invalidate-view table)
          (resize-superiors (ccl:view-container table)
                            0 (- (xy-y newsize) (xy-y oldsize))))))))

(defun redisplay-changed-table (table)
  (multiple-value-bind (new old) (size-table table 
                                             (ccl:view-size table))
    (let ((dy (xy-y (xy- new old)))
          (claims (table-claims table)))
      (if (or (= dy 0) claims)
        (progn (ccl:invalidate-view table)
               (and claims (update-scrollbar-visibility table))
               )
        (let ((window (ccl:view-window table))
              (super (ccl:view-container table)))
          (adjust-later-views window
                              (xy-y (ccl:convert-coordinates
                                     (ccl:view-position table)
                                     super window))
                              dy)
          (resize-superiors super 0 dy)))
      t)))

(defun update-scrollbar-visibility (view)
  (let ((window (ccl:view-window view))
        (scroll (ccl::v-scroller view)))
    (when window
      (if (<= (xy-y (table-extent view)) (xy-y (table-claims view)))
        (when (ccl:view-window scroll) 
          (ccl:remove-subviews window scroll))
        (unless (ccl:view-window scroll) 
          (ccl:add-subviews window scroll))))))

;;;
;;; Slot table displays an object's slots in a table. current slot values
;;; are shown in edit buffers.
;;;

(defclass slot-table (table object-view-mixin)
  ((block :initarg :block :initform nil :accessor slot-table-block)
   (slots :accessor slot-table-slots))
  (:default-initargs :view-nick-name ':slot-table) )

(defmethod shared-initialize :after ((view slot-table) slots &rest args)
  ;; creates table entries out of slots in the associated object. this 
  ;; code needs to execute BEFORE the table class lays out the table.
  (declare (ignore slots args))
  (let* ((object (slot-value view 'object))
         (block (slot-value view 'block))
         slots vals pars entries)
    (cond ((typep object 'standard-class)
           (setf slots (class-slots object))
           (setf vals nil))
          ((typep object 'standard-object)
           (setf slots (class-slots (class-of object)))
           (setf vals t))
          (t
           (error "~S not an object or class." object)))
    (setf slots (mapcar #'slot-definition-name slots))
    (setf (slot-value view 'slots) slots)
    (setf pars (object-parameters object))
    (setf entries
          (loop for slot in slots
                unless (or (member slot block)
                           (and pars
                                (not (find slot pars 
                                           :key #'parameter-name))))
                collect 
                (make-table-entry
                 :id slot
                 :label (string-capitalize (string slot))
                 :editor (if (and vals (slot-boundp object slot))
                           (slot-value object slot) nil)
                 :mode (logior +entry-hideable+ +entry-moveable+))))
    (setf (slot-value view 'entries) entries)
    view))

(defun inspect-slots (table &optional allslots?)
  (let ((object (slot-value table 'object))
        redraw)
    (if allslots?
      (if (table-hidden-entries? table)
        (progn
          (map-table #'unhide-table-entry table)
          (setf redraw t))
        (let ((object (object-view-object table))
              (block (slot-table-block table))
              (slots (slot-table-slots table))
              (mode (logior +entry-readonly+ +entry-hideable+
                            +entry-moveable+)))
          (loop for slot in slots
                unless (or (member slot block)
                           (table-find slot table :key #'table-entry-id))
                do (add-table-entry 
                    table 
                    (make-table-entry
                     :id slot :mode mode
                     :editor (careful-slot-value object slot))
                    nil)
                (setf redraw t))))
      (if (table-hidden-entries? table)
        nil
        (let ((pars (object-parameters object))) 
          (map-table #'(lambda (p) 
                         (unless (find (table-entry-id p) 
                                       pars :key #'parameter-name)
                           (hide-table-entry p)))
                     table)
          (setf redraw t))))
    (when redraw
      (multiple-value-bind (newsize oldsize)
                           (size-table table (ccl:view-size table))
        (if (or (xy= newsize oldsize) (table-claims table))
          (ccl:invalidate-view table)
          (resize-superiors (ccl:view-container table)
                            0 (- (xy-y newsize) (xy-y oldsize))))))))

(defun slot-table-revert (view)
  (let* ((object (slot-value view 'object))
         (values (if (typep object 'standard-object) t nil)))
    (map-table #'(lambda (e) 
                        (ccl:set-dialog-item-text
                         (table-entry-editor e)
                         (if (and values
                                  (slot-boundp object
                                               (table-entry-id e)))
                           (prin1-to-string
                            (slot-value object
                                        (table-entry-id e)))
                           "")))
                    view)))

;;;
;;;
;;;

(defclass table-label (label)
  ()
  (:default-initargs
    :framed? nil))

;(defclass table-label (ccl:view)
;  ((text :initform "" :initarg :label-text :accessor label-text
;         :initarg :DIALOG-ITEM-TEXT)
;   )
;  (:default-initargs :view-font *capella-font*))

;(defmethod ccl:view-draw-contents ((view table-label))
;  (draw-label view ))

;(defmethod ccl:dialog-item-text ((view table-label))
;  (slot-value view 'text))

;(defmethod ccl:set-dialog-item-text ((view table-label) text)
;  (setf (slot-value view 'text) text)
;  (ccl:invalidate-view view))

(defmethod table-label-string (label (table table))
  ;; label can return a different string for each type of table.
  (item-title label))

(defmethod hilite-view ((view table-label) &optional hilite enabled)
  (declare (ignore enabled))
  (let ((size (ccl:view-size view))
        (rise (ccl:font-info (ccl:view-font view)))
        (color *bgc*)) ;(table-entry-color (ccl:view-container view))
    (ccl:with-font-focused-view view
      (ccl:rlet ((rect :rect :topleft 0 :bottomright size))
        (ccl:with-fore-color (if hilite ccl:*black-color* color)
          (#_fillrect rect ccl:*black-pattern*))
        (ccl:with-fore-color (if hilite color ccl:*black-color*)
          (ccl:with-pstrs ((text (slot-value view 'text)))
            (#_moveto 0 rise)
            (#_drawstring text)))))))

(defmethod ccl:view-click-event-handler :around ((view table-entry) where)
  (if (ccl:control-key-p)
    (let ((sub (ccl:find-view-containing-point view where)))
      (if sub
        (if (eq sub (table-entry-label view))
          (table-label-click sub where)
          (call-next-method))
        (call-next-method)))
    (call-next-method)))

(defun table-label-click (label where)
  (let* ((entry (ccl:view-container label))
         (table (ccl:view-container entry))
         (window (ccl:view-window entry)))
    (cond ((ccl:control-key-p) 
           (if (ccl:double-click-p )
             (when (logtest (table-entry-mode entry) +entry-hideable+) 
               (loop while (ccl:mouse-down-p) do nil)
               (when (ccl:view-contains-point-p 
                      entry (ccl:view-mouse-position table))
                 (hide-table-entry entry)
                 (redisplay-changed-table table)))
             (when (logtest (table-entry-mode entry) +entry-moveable+)
               (let ((oldpos (ccl:view-position entry))
                     mouse other x y)
                 (setf x (xy-x oldpos))
                 (setf y (xy-y where))
                 (loop while (ccl:mouse-down-p)
                       do
                       (setf mouse (ccl:view-mouse-position table))
                       (ccl:set-view-position entry (xy x (- (xy-y mouse) y)))
                       (ccl:view-draw-contents entry)
                       (ccl:window-update-event-handler window))
                 (ccl:set-view-position entry oldpos)
                 (setf other (ccl:find-view-containing-point
                              table (xy-x mouse) (xy-y mouse) t))
                 (if (eq other entry)  ; didn't move
                   (progn (ccl:view-draw-contents entry) 
                          (ccl:window-update-event-handler window))
                   (let ((up? (< (xy-y mouse) (xy-y oldpos))))
                     (if (null other)  ; moved before or after visible entries
                       (move-table-entry table entry 
                                         (table-position-if 
                                          #'table-entry-visible? 
                                          table :from-end (not up?)))
                       (let ((pos (table-position other table)))
                         (move-table-entry
                          table entry
                          (if (< (xy-y (ccl:convert-coordinates 
                                        mouse table other))
                                 (floor (view-height other) 2)) 
                            (if up? pos (1- pos))
                            (if up? (1+ pos) pos)))))
                     (redisplay-changed-table table)))))))
          ((ccl:shift-key-p)
           ;; copy label text to current input buffer, if any
           (let ((buf (ccl:current-key-handler window)))
             (when buf
               (ccl:buffer-insert (ccl:fred-buffer buf) 
                                  (table-label-string label table))
               (ccl:fred-update buf))))
          
          )))

#|
;;; unused
(defmethod deselect-label (label)
  (when (label-selected? label)
    (setf (label-selected? label)  nil)
    (draw-label label nil)))

(defmethod select-label (label)
  (unless (label-selected? label)
    (setf (label-selected? label) t)
    (draw-label label t)))
|#




