;;; **********************************************************************
;;; Copyright (c) 89-94, 95 Heinrich Taube.  All rights reserved.
;;; Use and copying of this software and preparation of derivative works
;;; based upon this software are permitted and may be copied as long as 
;;; no fees or compensation are charged for use, copying, or accessing
;;; this software and all copies of this software include this copyright
;;; notice.  Suggestions, comments and bug reports are welcome.  Please 
;;; address email to: hkt@zkm.de
;;; **********************************************************************

(in-package :stella)


(defvar *object-table* (make-hash-table :test #'equal)
  "Table for objects with associated id.")

(defun map-objects (fn &optional (type t))
  (maphash #'(lambda (key val)
               (declare (ignore key))
               (when (typep val type)
                 (funcall fn val)))
           *object-table* ))

(defun find-object (id &optional errorp)
  (or (gethash id *object-table*)
      (and errorp (error "No element with id: ~S" id))))

(defun find-element (&rest args)	; need for backward compatibility
  (apply #'find-object args))

(defsetf find-object (id) (element)
  `(setf-find-object ,element ,id))

(defun setf-find-object (element id)
  (when (slot-exists-p element 'id)
    (setf (slot-value element 'id) id))
  (setf (gethash id *object-table*) element))

(defun list-all-objects (&key top-level system)
  (declare (ignore top-level system))
  (let ((list ()))
    (maphash #'(lambda (key val)
    	         (declare (ignore key))
                 (push val list))
  	     *object-table*)
    list))

(defun name (name &optional (new nil))
  (let ((sym (etypecase name (symbol name)(string (intern name)))))
    (if new (gentemp (symbol-name name)) sym)))

(defun object-namestring (obj &optional (pretty t) )
  (if obj
    (let ((name (object-name obj)))
      (if name
        (if (and pretty (typep obj 'id-mixin))
          (string-capitalize (string name))
          (if (stringp name) name (princ-to-string name)))
        (let ((sup (the-container obj nil)))
          (if sup
            (format nil "~:[~A~;~:(~A~)~][~D]"
                    pretty (object-name sup)
                    (1+ (object-position obj sup)))
            (format nil "~:[~A~;~:(~A~)~]"
                    pretty 
                    (class-name (if (typep obj 'standard-class)
                                  obj (class-of obj))))))))
    nil))

;;;
;;; clos utilities
;;;

(defun class-subclasses (class &optional include-self)
  (let ((kids (loop for kid in (class-direct-subclasses class)
                    nconc (class-subclasses kid t))))
    (if include-self (cons class kids) kids)))

(defun class-slot-p (class slot)
  (find slot (class-slots class) :key #'slot-definition-name))

(defun expand-initargs (class inits)
  #+aclpc (unless (class-finalized-p class) (finalize-inheritance class))
  (loop with slots = (class-slots class) 
        and saved = inits and arg and init and value
        while inits
        do (setf init (pop inits))
           (setf arg (if (constantp init) init `(quote ,init)))
           (if inits 
               (setf value (pop inits))
             (error "Uneven initialization list: ~S" saved))
        if (find init slots :key #'slot-definition-initargs :test #'member)	  
        nconc (list arg value) into parsed
        else 
        do (error "~S is not a legal initialization argument for ~A."
                  init (class-name class) )
        finally (return parsed)))
				 
(defmacro slot-value-or-default (instance slot &optional (default nil))
  (let ((i (gensym))
  	(s (gensym)))
   `(let ((,i ,instance)
          (,s ,slot))
       (if (slot-boundp ,i ,s) (slot-value ,i ,s) ,default))))

(defmacro careful-slot-value (obj slot &optional default)
  (let ((ov (gensym))(sv (gensym))(dv (gensym)))
    `(let ((,ov ,obj)(,sv ,slot)(,dv ,default))
       (and (slot-exists-p ,ov ,sv)
            (slot-boundp ,ov ,sv)
            (or (slot-value ,ov ,sv) ,dv)))))
	    
;;;
;;; a bag is a list structure optimized for appending ala interlisp's TCONC
;;; list constructor, and for cycling through the elements over and over again.
;;; a bag looks like:  (cache tail . current), where cache holds all the
;;; elements added since the bag was created, tail points to the last cons
;;; cell in the cache, and current are the current bag elements. popping
;;; elements removes them from current, not from the cache.
;;;
     
(defun make-bag (&optional elements (copy-first t)) 
  (when copy-first (setf elements (copy-list elements)))
  (list* elements (last elements) elements))

(defmacro bag-elements (bag)
  `(cddr ,bag))

(defmacro bag-tail (bag)
  `(cadr ,bag)) 

(defmacro bag-cache (bag)
  `(car ,bag))

(defmacro push-bag (x bag)
  ;; push new thing onto cache, don't update elements.
  (let ((var (gensym)))
    `(let ((,var ,bag))
       (push ,x (bag-cache ,var))
       (when (null (bag-tail ,var))
         (setf (bag-tail ,var) (bag-cache ,var)))
       ,var)))

(defmacro append-bag (x bag)
  ;; append new thing to cache, don't update elements.
  (let ((var (gensym))
        (new (gensym)))
    `(let ((,var ,bag)
           (,new (list ,x)))
       (if (null (bag-cache ,var))
           (setf (bag-cache ,var) (setf (bag-tail ,var) ,new))
         (setf (bag-tail ,var)
	       (cdr (rplacd (bag-tail ,var) ,new))))
       ,var)))

(defmacro insert-bag (element bag pos)
  (let ((ev (gensym))(pv (gensym))(bv (gensym))(tv (gensym)))
    `(let ((,ev ,element)
           (,pv ,pos)
	   (,bv ,bag))
       (if (or (null (bag-cache ,bv))
               (= ,pv 0))
           (push-bag ,ev ,bv)
         (if (< ,pv (length (bag-cache ,bv)))
             (let ((,tv (nthcdr (1- ,pv) (bag-cache ,bv))))
               (setf (cdr ,tv) (cons ,ev (cdr ,tv)))
	       ,bv)
	   (append-bag ,ev ,bv))))))
	   
(defmacro pop-bag (bag)
  (let ((var (gensym)))
    `(let ((,var ,bag))
       (pop (bag-elements ,var)))))
  
(defmacro reset-bag (bag)
  (let ((var (gensym)))
    `(let ((,var ,bag))
       (setf (bag-elements ,var) (bag-cache ,var))
       ,var)))
	
(defmacro forget-bag (bag)
  (let ((var (gensym)))
    `(let ((,var ,bag))
       (setf (bag-tail ,var) nil 
             (bag-cache ,var) nil
	     (bag-elements ,var) nil)
      ,var)))

;;;
;;; bags as resourced score queues...
;;;

(defparameter *scheduler-resource* nil) ; cons resourced for scheduling.

(eval-when (load eval)
  (defresource *scheduler-resource* nil :size 128 :constructor 
    #'(lambda (&rest args) args (cons nil nil)))
  (initialize-resource *scheduler-resource* 128))

(defmacro %rcons (object tail)
  (let ((cell (gensym)))
    `(let ((,cell (fast-allocate-resource *scheduler-resource*)))
       (setf (car ,cell) ,object (cdr ,cell) ,tail)
       ,cell)))

(defmacro %rpop (list)
  (let ((obj (gensym))
        (old (gensym)))
    `(let* ((,old ,list)
            (,obj (car ,old)))
       (when ,old 
         (setf ,list (cdr ,old) )
         (fast-deallocate-resource *scheduler-resource* nil ,old))
       ,obj)))

#|
;; these versions of %rcons and %rpop disable the scheduler's resource
;; and are only here for debugging. to use them you must recompile
;; this file, select.lisp and midi.lisp.
(defmacro %rcons (a b) `(cons ,a ,b))
(defmacro %rpop (list) `(pop ,list))
|#

; out of date, dont use.
;(defmacro %rpop (list)
;  (let ((obj (gensym))
;        (old (gensym)))
;    `(and ,list
;          (let* ((,old ,list)
;                 (,obj (car ,old)))
;            (setf  (car ,old) nil
;                 ,list (cdr ,old) ; (cdr ,list) 
;                   (cdr ,old) nil
;                  )
;            (fast-deallocate-resource *scheduler-resource* nil ,old)
;            ,obj))))

(defmacro bag-enqueue (element bag &optional (accessor 'object-time)
                               earliest)
  ;; accessor is object-time normally, but midifiles enqueue pending
  ;; note offs as cons cells (time . message)
  ;; if earliest is NIL then the object is placed at the last possible
  ;; time avaliable so that objects of the same time are run in the
  ;; order the user specified.  noteOff sheduling, on the other hand,
  ;; wants to place a future note off at the earliest time possible
  (let ((evar (gensym))
        (bvar (gensym)))
    `(let ((,evar ,element)
           (,bvar ,bag))
       (cond ((null (bag-elements ,bvar))
              ;(push ,evar (bag-elements ,bvar))
              (setf (bag-elements ,bvar) 
                    (%rcons ,evar (bag-elements ,bvar)))
              (setf (bag-tail ,bvar) (bag-elements ,bvar)))
             (t 
	      (let ((time (,accessor ,evar)))
                (declare (number time) (optimize (speed 3) (safety 0)))
                (cond ((, (if earliest '> '>=) 
                        time 
                        (the number (,accessor (car (bag-tail ,bvar)))))
                       (setf (bag-tail ,bvar)
	                     (cdr (rplacd (bag-tail ,bvar)
                                          (%rcons ,evar nil)))))
                      ((, (if earliest '<= '< )
                        time 
                        (the number
                          (,accessor (car (bag-elements ,bvar)))))
                       (setf (bag-elements ,bvar) 
                             (%rcons ,evar (bag-elements ,bvar))))	       
                      (t 
                       (loop with tail = (bag-elements ,bvar)
                             for head on (cdr tail)
                             while
                             (,(if earliest '< '<= )
                              (the number (,accessor (car head)))
                                 time)
                             do (setf tail head)
                             finally (rplacd tail 
                                             (%rcons ,evar head))))))))
       ,evar)))

(defmacro bag-dequeue (bag)
  `(%rpop (bag-elements ,bag)))

#|
;; resource test.
(progn
  (defparameter %cells (loop for i below 32 collect (cons nil nil)))
  nil)


(defmacro %rcons (object list)
  (let ((cell (gensym)))
    `(let ((,cell (nextfree)))
       (setf (car ,cell) ,object (cdr ,cell) ,list)
       ,cell)))

(defmacro %rpop (list)
  (let ((obj (gensym)) (old (gensym)))
    `(and ,list
          (let* ((,old ,list) (,obj (car ,old)))
            (setf (car ,old) nil ,list (cdr ,list) (cdr ,old) nil)
            ,obj))))

(defun nextfree ()
  (or (loop for i in %cells unless (car i) return i)
      (loop for i below 32
            do (push (cons nil nil) %cells)
            finally (return (car %cells)))))

(defmacro bag-enqueue (element bag &optional (accessor 'object-time))
  (let ((evar (gensym))
        (bvar (gensym)))
    `(let ((,evar ,element)
           (,bvar ,bag))
       (cond ((null (bag-elements ,bvar))
              ;(push ,evar (bag-elements ,bvar))
              (setf (bag-elements ,bvar) (%rcons ,evar (bag-elements ,bvar)))
              (setf (bag-tail ,bvar) (bag-elements ,bvar)))
             (t 
	      (let ((time (,accessor ,evar)))
                (declare (number time) (optimize (speed 3) (safety 0)))
                (cond ((>= time 
                           (the number (,accessor (car (bag-tail ,bvar)))))
                       (setf (bag-tail ,bvar)
	                     (cdr (rplacd (bag-tail ,bvar)
                                          (%rcons ,evar nil)))))
                      ((< time 
                          (the number (,accessor (car (bag-elements ,bvar)))))
                       (setf (bag-elements ,bvar) 
                             (%rcons ,evar (bag-elements ,bvar))))	       
                      (t 
                       (loop with tail = (bag-elements ,bvar)
                             for head on (cdr tail)
                             while (<= (the number (,accessor (car head)))
                                       time)
                             do (setf tail head)
                             finally (rplacd tail (%rcons ,evar head))))))))
       ,evar)))

(defmacro bag-dequeue (bag)
  `(%rpop (bag-elements ,bag)))
|#


#|
original:
(defmacro bag-enqueue (element bag &optional (accessor 'object-time))
  (let ((evar (gensym))
        (bvar (gensym)))
    `(let ((,evar ,element)
           (,bvar ,bag))
       (cond ((null (bag-elements ,bvar))
              (push ,evar (bag-elements ,bvar))
              (setf (bag-tail ,bvar) (bag-elements ,bvar)))
             (t 
	      (let ((time (,accessor ,evar)))
                (declare (float time) (optimize (speed 3) (safety 0)))
                (cond ((>= time 
                           (the float (,accessor (car (bag-tail ,bvar)))))
                       (setf (bag-tail ,bvar)
	                     (cdr (rplacd (bag-tail ,bvar) (list ,evar)))))
                      ((< time 
                          (the float (,accessor (car (bag-elements ,bvar)))))
                       (push ,evar (bag-elements ,bvar)))	       
                      (t 
                       (loop with tail = (bag-elements ,bvar)
                             for head on (cdr tail)
                             while (<= (the float (,accessor (car head)))
                                       time)
                             do (setf tail head)
                             finally (rplacd tail (cons ,evar head))))))))
      ,evar)))

(defmacro bag-dequeue (bag)
  `(pop-bag ,bag))
|#
  
;;;
;;; readtable hackery.
;;;
  
(defun find-object-reader (stream subchar arg)
  (declare (ignore subchar arg))
  (let ((thing (read stream t nil t)))
    (values `(find-object ',thing t))))
    
(set-dispatch-macro-character #\# #\! #'find-object-reader 
                              cm::*common-music-readtable*)

(defun expr-reader (stream subchar arg)
  (declare (ignore subchar arg))
  (let ((form (read stream t nil t)))
    (values `(expr ,form))))

(set-dispatch-macro-character #\# #\e #'expr-reader 
                              cm::*common-music-readtable*)

;;;
;;; instance printing - stolen from pcl
;;;

#-pcl
(defmacro printing-random-thing ((thing stream) &body body)
  (let ( (s (gensym)))
    `(let ((,s ,stream))
       (format ,s "#<") 
       ,@body
       (format ,s "~@[#x~X~]>" (address-of ,thing)))))

;;;
;;; defobject
;;;

(defvar *defobjects* (make-hash-table)
  "Defobject class information.")

(defmacro defobject (name supers slots &rest options)
  (let (classoptions stellaforms pars)
    (dolist (o options)
      (case (first o)
        (:parameters
         (setf pars (cdr o)))
        (t
         (push o classoptions))))
    (push `(define-object-parameters (find-class ',name)
             ',(or pars t))
          stellaforms)
    #+(and excl cltl2)
    (push `(clos:finalize-inheritance (find-class ',name)) stellaforms)
    `(progn
       (defclass ,name ,supers 
         ,(loop for x in slots
                when (consp x)
                collect
             (let* ((inits (copy-list x))
                    (slot (first inits))
                    (keyword (intern (string slot) :keyword))
                    sym? key?)
               (loop for (key val) on (cdr inits) by #'cddr
                     do 
                 (when (eq key ':initarg) 
                   (if (eq val slot) (setf sym? t)
                     (if (eq val keyword) (setf key? t)))))
               (unless sym?
                 (push slot (cdr inits)) 
                 (push ':initarg (cdr inits)))
               (unless key?
                 (push keyword (cdr inits))
                 (push ':initarg (cdr inits)))
               inits)
                else 
                collect (list x ':initarg x 
                              ':initarg (intern (string x) :keyword)))
         ,@classoptions)
       ,@stellaforms
       (find-class ',name))))

;;;
;;; object parameter declarations are stored in a hash table. in
;;; the best of all worlds I would use the MOP to define a
;;; class-with-parameters and a parameter-definition class, but many
;;; of the CLOS MOP implementations are still too weak for this.
;;;

(defstruct (parameter (:type list))
  name)

(defmethod object-parameters ((object standard-class))
  (gethash object *defobjects*))

(defmethod object-parameters ((object t))
  (gethash (class-of object) *defobjects*))

(defun define-object-parameters (class &optional (specs t))
  (flet ((canonicalize-spec (spec)
           (if (symbolp spec)
             (list spec)
             spec)))
    (cond ((eq specs t)
           (setf (gethash class *defobjects*)           
                 (loop for super in (class-precedence-list class)
                       thereis (object-parameters super))))
          (t
           #+(and excl cltl2) (clos:finalize-inheritance class)
           #+aclpc (unless (class-finalized-p class)
                     (finalize-inheritance class))
           (let ((internal (mapcar #'canonicalize-spec 
                                   specs))
                 (slots (class-slots class)))
             (dolist (spec internal)
               (unless (find (first spec)
                             slots :key #'slot-definition-name)
                 (error "Attempt to define ~A as a parameter ~
                         but ~A has no such slot." 
                        (first spec) (class-name class))))
             (setf (gethash class *defobjects*)           
                   internal))))))

(defun map-classes (fn &optional (type t))
  (maphash #'(lambda (key val)
               (declare (ignore val))
               (when (typep (class-prototype key) type)
                 (funcall fn key)))
           *defobjects*))

(defun element-classes (&optional mode)
  (declare (ignore mode))
  (let ((list ()))
    (map-classes #'(lambda (o) (push o list)) 'element)
    (nreverse list)))

(defun container-classes (&optional mode)
  (declare (ignore mode))
  (let ((list ()))
    (map-classes #'(lambda (o) (push o list)) 'startable-element)
    (nreverse list)))



