;;;
;;;
;;; File:         pmcl-midi.lisp
;;;
;;; Summary:      LISP part of the pmclmidi shared-library reimplementation
;;;               of the Apple MidiManager interface for the PPC version of 
;;;               Macintosh Common Lisp (3.9 or higher).
;;;
;;; Dependencies: pmcl-midi.cc
;;;               pmcl-midi.hh
;;;               pmcl-entry.lisp
;;;
;;;               ***********************************************************
;;;               Copyright (c) 1996 by Tobias Kunze and Rick Taube.  All
;;;               rights reserved.  Use and copying of this software and
;;;               preparation of derivative works based upon this software
;;;               are permitted and may be copied as long as no fees or
;;;               compensation are charged for use, copying, or accessing
;;;               this software and all copies of this software include this
;;;               copyright notice.  Suggestions, comments and bug reports
;;;               are welcome.  Please address email regarding this midi
;;;               implementation to: t@kunze.stanford.edu or
;;;               hkt@ccrma.stanford.edu.
;;;               ***********************************************************
;;;
;;; Author:       Tobias Kunze
;;; e-Mail:       t@kunze.stanford.edu
;;; Org:          CCRMA, Stanford University
;;;
;;; orig-Date:    20-May-96 at 02:31:36
;;; Last-Mod:     29-Jul-96 at 14:56:02 by Tobias Kunze
;;;
;;;
;;; Description:  This code, together with the files pmcl-midi.cc, 
;;;               pmcl-midi.hh and pmcl-entry.lisp interfaces with the Apple 
;;;               Midi Manager.  It is intended to be part of Common Music by
;;;               Rick Taube but may be used in standalone applications as  
;;;               well. To use it without Common Music, follow the 
;;;               instructions in the `CLIENT DEFINITIONS' sections in this
;;;               file below and in pmcl-entry.lisp.
;;;
;;;               To customize its behaviour or learn about user-definable
;;;               parameters, see section USER VARIABLES below.
;;;
;;;
;;;               OVERVIEW
;;;
;;;               o  FUNCTIONALITY:  this code implements the following
;;;                  functionality:  
;;;
;;;                     NOTE: function names that are meant to be accessed by
;;;                           clients start with a '%midi-' prefix.  DO NOT
;;;                           CALL ANY OTHER FUNCTION UNLESS YOU KNOW EXACTLY
;;;                           WHAT YOU ARE DOING.  See FUNCTION SWAPPING
;;;                           below. 
;;;
;;;                     -----------------------------------------------------
;;;                     %midi-open <port>                          [Function]
;;;                             
;;;                     Open midi.  <port> may be 0, :a, or :modem to specify
;;;                     serial port A; 1, :b, or :printer to specify port B;
;;;                     2, :a+b, or :both to specify both serial ports or -1
;;;                     or nil to leave it to the implementation.
;;;
;;;                     -----------------------------------------------------
;;;                     %midi-close                                [Function]
;;;
;;;                     Close midi and free all allocated memory.
;;;
;;;                     -----------------------------------------------------
;;;                     %midi-read-messages                        [Function]
;;;
;;;                     Poll the MidiManager to read all currently pending
;;;                     messages.  The MidiManager calls whatever function is
;;;                     held in the global *midi-read-hook* (see the section
;;;                     CLIENT DEFINITIONS below).
;;;
;;;                     -----------------------------------------------------
;;;                     %midi-write-message <message> <time>       [Function]
;;;
;;;                     Write the message <message> at time <time>.
;;;
;;;                     -----------------------------------------------------
;;;                     %midi-all-notes-off <time>                 [Function]
;;;
;;;                     Transmits note-off messages for every key number on
;;;                     every channel on both output ports.  The time 
;;;                     parameter is ignored.
;;;
;;;                     -----------------------------------------------------
;;;                     %midi-flush-transmit                       [Function]
;;;
;;;                     Flush the output queues.
;;;
;;;                     -----------------------------------------------------
;;;                     %midi-flush-receive                        [Function]
;;;
;;;                     Flushes pending input.
;;;
;;;                     -----------------------------------------------------
;;;                     %midi-hush                                 [Function]
;;;
;;;                     Flush both input and output queues.
;;;
;;;                     -----------------------------------------------------
;;;                     %midi-start-time                           [Function]
;;;
;;;                     Start the Midi timer.  Midi time is measured in ms up
;;;                     from 0. 
;;;
;;;                     -----------------------------------------------------
;;;                     %midi-stop-time                            [Function]
;;;
;;;                     Stop the Midi timer.
;;;
;;;                     -----------------------------------------------------
;;;                     %midi-get-time                             [Function]
;;;
;;;                     Get the current Midi time in ms.
;;;
;;;                     -----------------------------------------------------
;;;                     %midi-set-time <time>                      [Function]
;;;                     
;;;                     Set the current Midi time in ms.
;;;                    
;;;                     ----------------------------------------------------- 
;;;                     
;;;                  
;;;               o  32-CHANNEL ENCODING:  channel messages with bit 15 (the
;;;                  most significant bit of data byte 2) set are routed to
;;;                  outport B instead of outport A and messages arriving at
;;;                  inport B get bit 15 set.
;;;               o  OUTPUT BUFFERING:  all messages are sent to an invisible
;;;                  output port, which directs them to a dynamic set of
;;;                  invisible input ports for queuing and time sorting.
;;;                  When a message becomes current, the MidiManager calls
;;;                  the invisible input port's read hook, which passes it on
;;;                  to either of the two output ports, depending on their
;;;                  "port" bit (bit 15).  New invisible input ports are
;;;                  allocated when all existing queues are filled and
;;;                  deallocated when they have been idling for a given
;;;                  time. If all queues are filled and if there is no more
;;;                  room for new ports, the output routine blocks.
;;;               o  FLUSHING:  the read hook for invisible input ports may be
;;;                  set to "flushing mode", in which case it dumps all
;;;                  pending note-off and note-on events with zero velocity.
;;;                  Since the MidiManager doesn't provide a means to test
;;;                  whether a port is empty, however, the mode is reset
;;;                  automatically after a given time interval (see USER
;;;                  VARIABLES below)---whether the queues are empty or not.
;;;                  Set this time interval to match your needs.
;;;               o  FUNCTION SWAPPING:  calling MidiManager calls at improper
;;;                  times crashes the Macintosh OS.  To protect users from
;;;                  calling such an improper trap accidentally, this code
;;;                  implements a function swapping mechanism that replaces
;;;                  client functions with an error-signaling stub function
;;;                  at improper times.
;;;
;;;
;;;               LIMITATIONS
;;;
;;;               o  Apple's MidiManager limits a client to 16 ports.  As a
;;;                  result, I can't allocate more than 10 invisible input
;;;                  ports, which do the actual output queuing and sorting (6
;;;                  ports: the time port, two input and two output ports and
;;;                  the invisible output port are always allocated).  Since
;;;                  a port's queue has a maximum buffer size of #xFFFF and
;;;                  since the Midi Manager allocates 16 bytes per message,
;;;                  this limits the number of possible events that may be
;;;                  scheduled for output at any given time to 10 * 4096.
;;;               o  The c sniffer routine cannot remove ports at the time it
;;;                  detects they are idle, since it runs at interrupt time,
;;;                  which is why I call a separate function from Lisp at
;;;                  non-interrupt time---currently whenever a midi message
;;;                  is written to the queues.  It may thus be possible that
;;;                  one or more ports with empty queues have been marked as
;;;                  such but not yet actually deallocated until another midi
;;;                  events is being written or midi closed.
;;;
;;;
;;;               KNOWN PROBLEMS
;;;
;;;               o  None. Honestly.
;;;
;;;
;;; Revision:     1.2
;;;
;;; Changes:
;;;   07-Jul-96   tk   Revision 1.0
;;;   12-Jul-96   tk   fixed bug that error'ed when the blessed driver had no 
;;;                    open ports, added QuickTime Midi Player support
;;;   18-Jul-96   hkt  ccl:define-entry-points moved to pmcl-entry.lisp to 
;;;                    insure that these get defined before this file is 
;;;                    compiled. 
;;;   28-Jul-96   tk   removed *midi-multitimbral-channels*
;;;


;;; -------------------------------------------------------------------------
;;;
;;; CLIENT DEFINITIONS
;;;

;;;
;;;      Our package.

(in-package #+cm :common-music #-cm :cl-user)

;;;
;;;   +client-name+                  <default: #+cm (cm-version) #-cm "LISP">
;;;      The name of our client as it appears in PatchBay.

(defconstant +client-name+ #+cm (cm-version nil) #-cm "LISP")

;;;
;;;   +client-id+                          <default: #+cm :|CM  | #-cm :LISP>
;;;      The four-letter ID (OSType) of our client.

(defconstant +client-id+ #+cm :|CM  | #-cm :LISP)

;;;
;;;   +midi-success+                                             <default: 0>
;;;      Result code for no Error

#-cm
(defconstant +midi-success+ 0)

;;;
;;;
;;;   *midi-read-hook*                                         <default: nil>
;;;      The function to call for incoming messages.  This should be a
;;;      function object with the syntax 
;;;       
;;;         <function-name> message time
;;;
;;;      Do not set it if you are using Common Music, since it deals with it
;;;      internally. 

#-cm
(defvar *midi-read-hook* nil
  "The function to call for incoming messages.")


;;;
;;; END CLIENT DEFINITIONS
;;;
;;; -------------------------------------------------------------------------




;;; -------------------------------------------------------------------------
;;;
;;; USER VARIABLES
;;;

;;; 
;;;   *midi-preferences*     <default: '((:apple :oms :mtp :midiqt) (:a :b))>
;;;      A list of the known midi drivers and ports, sorted in descending
;;;      order according to their relative preference.
;;;      Note that the MidiQT extension and QuickTime Midi Player share a 
;;;      keyword, i.e. :midiqt.
;;;      The variable's primary use is to provide midi-open with the means
;;;      to discriminate between multiple possibilities, whenever it it is
;;;      called withoput parameters (as done by the midi menu's open item).

(defparameter *midi-preferences* '((:apple :mtp :oms :midiqt) (:a :b))
  "A list of known midi drivers and ports, sorted in descending ~
   order according to their relative preference.")

;;;
;;;   *midi-flush-timeout*	                              <default: 3000>
;;;      The time in milliseconds, for which the queueing mechanism stays in
;;;      flushing mode, therefore ignoring newly scheduled events.
;;;
;;;      Note: since flushing is achieved by dumping all pending note-off
;;;            messages,it  is possible for _huge_ amounts of buffered output
;;;            to leave messages unprocessed. This may lead to hanging notes,
;;;            if _very_ long notes that are already turned on happen to have
;;;            their note-off message among this unprocessed chunk of midi
;;;            data. Therefore, you may want to increase or decrease the
;;;            value of *midi-flush-timeout*, depending on the overall
;;;            musical structure you tend to produce. Note, too, that this
;;;            parameter is semi-global only. That is, any changes to it come
;;;            into effect only after the next time you open a midi port. 

(defparameter *midi-flush-timeout* 3000
  "Time interval (in ms) after which output flushing returns to normal mode.")

;;;
;;; END USER VARIABLES.
;;;
;;; -------------------------------------------------------------------------


;;;
;;; END DEFINITIONS
;;;



(export
 '(*midi-preferences* *midi-flush-timeout*
		      #-cm *midi-read-hook* #-cm +midi-success+))



;;; -------------------------------------------------------------------------
;;;
;;; Midi Menu
;;;

(defvar %midi-menu% nil)
(defvar %midi-open% nil)

(defconstant +open-name+ "Open")
(defconstant +close-name+ "Close")
(defconstant +hush-name+ "Hush")
(defconstant +flush-input-name+ "Flush Input")
(defconstant +flush-output-name+ "Flush Output")
(defconstant +notes-off-name+ "All Notes Off")
(defconstant +test-name+ "Test Midi")

(eval-when (:load-toplevel :execute)
  ;; make sure we close midi when exiting 
  (let ((quit (ccl:find-menu-item ccl:*file-menu* "Quit")))
    (defmethod ccl:menu-item-action :before ((i (eql quit))) 
      (when %midi-open%  (%midi-close))))
  ;; make a midi menu
  (unless (ccl:find-menu "Midi")
    (ccl:menu-deinstall ccl:*windows-menu*)
    (ccl:menu-install
     (setf %midi-menu%
           (make-instance 'ccl:menu
             :menu-title "Midi"
             :help-spec "Some useful midi commands."
             :menu-items 
             (list (make-instance 'ccl:menu-item 
                     :menu-item-title (if %midi-open% 
                                        +close-name+ +open-name+)
                     :menu-item-action 
                     #'(lambda ()
                         (ccl:eval-enqueue
                          '(when ccl:*top-listener*
                             (print
			      (if %midi-open%
				  (progn (%midi-close) t)
				(progn (%midi-open nil) %midi-open%))))))
                     :command-key #\m
                     :disabled nil
                     :help-spec "Open/close MIDI.")
                   (make-instance 'ccl:menu-item :menu-item-title "-")
                   (make-instance 'ccl:menu-item 
                     :menu-item-title +hush-name+
                     :disabled (if %midi-open%  nil t) 
                     :command-key #\;
                     :menu-item-action #'(lambda () (%midi-hush))
                     :help-spec
		     "Stops sound and flushes MIDI input and output queues.")
                   (make-instance 'ccl:menu-item 
                     :menu-item-title +flush-input-name+
                     :disabled (if %midi-open%  nil t) 
                     :command-key #\[
                     :menu-item-action #'(lambda () (%midi-flush-receive))
                     :help-spec "Flushes MIDI input queue.")
                   (make-instance 'ccl:menu-item 
                     :menu-item-title +flush-output-name+
                     :disabled (if %midi-open%  nil t) 
                     :command-key #\]
                     :menu-item-action #'(lambda () (%midi-flush-transmit))
                     :help-spec "Flushes MIDI output queue.")
                   (make-instance 'ccl:menu-item 
                     :menu-item-title +notes-off-name+
                     :disabled (if %midi-open%  nil t) 
                     :command-key #\'
                     :menu-item-action #'(lambda () (%midi-all-notes-off 0))
                     :help-spec "Stops all sounding notes.")
                   (make-instance 'ccl:menu-item :menu-item-title "-")
                   (make-instance 'ccl:menu-item 
                     :menu-item-title +test-name+
                     :disabled (if %midi-open%  nil t) 
                     :command-key #\`
                     :menu-item-action 
                     #'(lambda () 
                         (loop for x below 12
                               with time = (midi-get-time)
                               for note = (+ x (* 12 (+ 3 (random 5))))
                               do 
                               (%midi-write-message 
                                (make-note-on 0 note (+ 32 (random 64))) 
                                time)
                               (%midi-write-message 
                                (make-note-off 0 note (random 128)) 
                                (+ time (* (expt (+ 2 (random 6)) 2) 30)))
                               (incf time (* (expt (random 5) 2) 30))))
                     :help-spec "Sends a small number of random notes.")
                   #+capella
                   (make-instance 'ccl:menu-item
                              :menu-item-title "Play..."
                              :disabled (not %midi-open%) 
                              :menu-item-action 
                              #'(lambda () (worksheet 'midi-play)))
                   #+capella
                   (make-instance 'ccl:menu-item
                     :menu-item-title "Receive..."
                     :disabled (not %midi-open%) 
                     :menu-item-action 
                     #'(lambda () (worksheet 'midi-receive)))
                   ))))
    (ccl:menu-install ccl:*windows-menu*)))


(defun %update-midi-menu ()
  (ccl:set-menu-item-title
   (or (ccl:find-menu-item %midi-menu% +open-name+)
       (ccl:find-menu-item %midi-menu% +close-name+))
   (if %midi-open% +close-name+ +open-name+))
  (loop for s in (list +hush-name+ +flush-input-name+ +flush-output-name+
                       +notes-off-name+ +test-name+
                       #+capella "Play..." #+capella "Receive...")
        for i = (ccl:find-menu-item %midi-menu% s)
        do (if %midi-open% (ccl:menu-item-enable i)
             (ccl:menu-item-disable i))))

(defun set-open/close-menu-item-function (open-function close-function)
  (ccl:set-menu-item-action-function
   (find (if %midi-open% +close-name+ +open-name+)
         (ccl:menu-items cm::%midi-menu%)
         :key #'ccl:menu-item-title :test #'string-equal)
   #'(lambda ()
       (let ((openfn open-function)
             (closefn close-function))
         (ccl:eval-enqueue
          `(when ccl:*top-listener*
             (print
              (if %midi-open%
                 (funcall ,closefn)
                 (funcall ,openfn)))))))))


#+cm 
(defun %midi-default-channel-map (port)
  (declare (ignore port))
  (default-midi-channel-map))




;;; -------------------------------------------------------------------------
;;;
;;; MidiManager Utilities
;;;

(defmacro with-midi-checked (form)
  (let ((var (gensym)))
    `(let ((,var ,form))
       (cond ((or (null ,var) (= ,var +midi-success+)) 
              +midi-success+)
             ((member ,var '(#$MIDIVConnectErr #$MIDIVConnectMade 
                             #$MIDIVConnectRmvd))
              +midi-success+)
             ((= ,var #$MIDINoClientErr)
              (error "MIDI: no client with that ID found."))
             ((= ,var #$MIDINoPortErr)
              (error "MIDI: no port with that ID found."))
             ((= ,var #$MIDITooManyPortsErr)
              (error "MIDI: too many ports already installed in the system."))
             ((= ,var #$MIDITooManyConsErr)
              (error "MIDI: too many connections made."))
             ((= ,var #$MIDINoConErr)
              (error "MIDI: connection does not exist."))
             ((= ,var #$MIDIWriteErr)
              (error "MIDI: MIDIWritePacket couldn't write to all ~
                      connected ports."))
             ((= ,var #$MIDINameLenErr)
              (error "MIDI: name supplied is longer than 31 characters."))
             ((= ,var #$MIDIDupIDErr)
              (error "MIDI: duplicate client ID."))
             ((= ,var #$MIDIInvalidCmdErr)
              (error "MIDI: command not supported for port type."))
             ((= ,var #$memFullErr)
              (error "MidiManager: Not enough room in the mac heap zone."))
             (t
              (error "MIDI: failed with ~s" ,var))))))

(defmacro %mm-get-midiidlist (midimanager-call)
  `(let* ((hdl ,midimanager-call) list)
     (if (eq hdl nil) (error "There is not enough room to get client or~%~
                              port information")
         (setf list (loop for i from 0 
                          repeat (ccl:href hdl MidiIDList.numIDs) 
                          collect (ccl:href hdl (MidiIDList.list i)))))
     (ccl:require-trap #_disposhandle hdl)
     list))






;;; -------------------------------------------------------------------------
;;;
;;; Ports and Drivers
;;;

;;; 
;;; To comply with Common Music, this is an array of lists of portnames in the 
;;; order (<long-name> <short-name> <number> &rest more-specs)

(defvar %portrefs%
  #((:modem :a 0) 
    (:printer :b 1) 
    (:modem+printer :a+b 2 :both) 
    (:printer+modem :b+a 3)))

#+cm
(eval-when (:load-toplevel :execute)
  (setf *port-refs* %portrefs%))

(defun %translate-portref (ref result)
  (let ((p (find ref %portrefs% :test #'member)))
      (if p
        (funcall (ecase result
		   (:long #'first)
		   (:short #'second)
		   (:number #'third))
                 p)
        (error "~a is not a valid port reference." ref))))


(defconstant +buffer-size+ #xFFFF)	; roughly 4096 messages (the maximum)

;;;
;;; These two constants are missing in the MIDI.lisp interface file.

(defconstant $midiInvisiblePort #x8000) ; this IS in the C interface
(defconstant $midiPortTypeMask #x0007)


(defstruct port
  name
  type					; Time, Input or Output, visible or not
  id 
  (format  #$midiFormatBeats)		; matters only for Time ports
  (offset #$midiGetCurrent)		; matters only for Input ports
  (refnum nil))				; reference number

(defstruct client
  name                                  ; Name as it appears in PatchBay
  id					; MidiManager Client ID
  time-port				; time base
  inport-a				; IN channels 0-15, port A
  inport-b				; IN channels 0-15, port B
  outport-a				; OUT channels 0-15, port A
  outport-b				; OUT channels 0-15, port B
  inv-outport				; Lisp OUT (C IN)
  (static-ports nil :type list) 	; list of above ports
  (dynamic-ports nil :type list))	; dynamic invisible input queues

(defstruct driver
  name
  id					; MidiManager Client ID
  filename
  ;; Can we connect to it automatically?
  can-connect				; t, :in or :out
  ;; These are in the order: time, in (from PatchBay), out (to PatchBay)
  a-ids					; ID's for serial port A
  b-ids)				; ID's for serial port B


;;;
;;; Our static ports.  Refnums and dynamic ports are added later.

(defvar %client%
  (let ((staticports (list (make-port :name ':|Time|
				       :type #$midiPortTypeTime
				       :id ':|Time|
				       :format #$midiFormatMSec)
			    (make-port :name ':|InputA|
				       :type #$midiPortTypeInput
				       :id ':|InA |
				       :offset #$midiGetNothing)
			    (make-port :name ':|InputB|
				       :type #$midiPortTypeInput
				       :id ':|InB |
				       :offset #$midiGetNothing)
			    (make-port :name ':|OutputA|
				       :type #$midiPortTypeOutput
				       :id ':|OutA|)
			    (make-port :name ':|OutputB|
				       :type #$midiPortTypeOutput
				       :id ':|OutB|)
			    (make-port :name ':|InvisibleOutput|
				       :type (logior #$midiPortTypeOutput 
						     $midiInvisiblePort)
				       :id ':|InvO|))))
    (make-client :name	       +client-name+
		 :id           +client-id+
		 :static-ports staticports
		 :time-port    (pop staticports)
		 :inport-A     (pop staticports)
		 :inport-B     (pop staticports)
		 :outport-A    (pop staticports)
		 :outport-B    (pop staticports)
		 :inv-outport  (pop staticports))))


;;;
;;; This constants are used to optimize midi output

(defvar +inv-outport+ (client-inv-outport %client%))
(defvar +client-inv-outport-id+ (port-id +inv-outport+))
(defvar +inv-outport-refnum+ nil)       ; pseudo-constant set by %midi-open

(defvar %driver-ports% nil)		; driver port ID's to connect to
					; 
(defvar %driver% nil)			; currently, Apple, Opcode, MidiQT, or 
                                        ;   Midi TimePiece. Initialized in
                                        ;   %mm-get-driver-and-ports


;;; Note OMS v. 1.2.3:
;;; portID's apparently read as: "opcode2midi" or "midi2opcode"
;;; "t" is time and 0-9 (8?) are ports.  port ":" is invisible...
;;; we connect to all by default.  ports are the same for a and b.
;;; a typical port list looks like:
;;; (:|O2mt| :|m2ot| :|O2m2| :|m2o:| :|O2m0| :|m2o0|)

#|
(defun %mm-get-port-info (clientID &rest portIDs)
  (loop for p in portIDs
        with hdl
        do (setf hdl (#_midigetportinfo clientID p))
        collect (ccl::href hdl Midiportinfo.portType)
        do (#_disposhandle hdl)))
;(%mm-get-port-info :|OmMd| :|O2mt| :|O2m2| :|m2ot| :|m2o:| :|O2m0| :|m2o0|)
|#

(defun %oms-ports (type)
  (declare (special %mm-gestalt%))
  (when (eq %mm-gestalt% :mmCMSetUp)
    (setf type (ecase type (:output "O2m") (:input "m2o")))
    (remove-if-not #'(lambda (x) 
                       (let ((str (symbol-name x)))
                         (and (string= str type :end1 2 :end2 2)
                              (digit-char-p (elt str 3)))))
                   (%mm-get-midiidlist (#_midigetports :|OmMd|)))))

;(%oms-ports :output)



;;; 
;;; The collection of known drivers.  Others may be added in the future.

(defvar %drivers%
  (list (make-driver :name ':apple
		     :filename "Apple MIDI Driver"
		     :id :|amdr|
		     :can-connect t
		     :a-ids '(:|ATim| :|Ain | :|Aout|)
		     :b-ids '(:|BTim| :|Bin | :|Bout|))
	;; OMS doesn't seem to distinguish serial ports (?)
	(let ((ids `(:|O2mt|		; or else :|m2ot|...
		       ,#'(lambda () (%oms-ports :output))
		       ,#'(lambda () (%oms-ports :input)))))
	  (make-driver :name ':oms
		       :filename "OMS MIDI Manager Driver"
		       :id :|OmMd|
		       :can-connect t
		       :a-ids ids
		       :b-ids ids))
	(make-driver :name ':mtp
		     :filename "MTP Driver 2.0"
		     :id :|MTPd|
		     :can-connect t
		     :a-ids '(:|timA| :MTPA :|Mod |
				:|Mod!| :|Mod\"| :|Mod\#| :|Mod$| 
				:|Mod%| :|Mod&| :|Mod'|)
		     :b-ids '(:|timB| :MTPB :|Prn | 
				:|Prn!| :|Prn\"| :|Prn\#| :|Prn$| 
				:|Prn%| :|Prn&| :|Prn'|))
        ;; (QuickTime Midi Player is actually the same as MidiQT)
	;; a and b ports are the same for MidiQT. no output ports.
        ;; :|in| has two #\Null's before the `i'!!
	(let ((ids (list :|time|
                         nil
                         (intern (format nil "~c~c~c~c" 
                                         #\Null #\Null #\i #\n)
                                 'keyword))))
	  (make-driver :name ':midiQT
		       :filename "MidiQT"
		       :id :|inpu|
		       :can-connect :in
		       :a-ids ids
		       :b-ids ids))))


;;;
;;; Utility accessors

(defun find-driver-by-name (ref)
  (or (find ref %drivers% :key #'driver-name)
      (error "Can't find driver named ~S." ref)))

(defun find-driver-by-id (ref)
  (or (find ref %drivers% :key #'driver-id)
      (error "Can't find driver with id ~S." ref)))

(defun find-driver-ports (num &key name id)
  (unless (numberp num) (error "~a is not a number" num))
  (let* ((driver (cond (name (find-driver-by-name name))
		       (id (find-driver-by-id id))
		       (t (error "need a client's name or ostype" ))))
	 (ports (case num
		  (0 (driver-a-ids driver))
		  (1 (driver-b-ids driver)))))
    (mapl #'(lambda (cns)
					;(describe (car cns))
	      (when (functionp (car cns))
		(setf (car cns) (funcall (car cns)))))
	  ports)))

;; The name confusion between portids and their access functions is not a bug:
;; think of it simply as icons appearing in Patch-Bay; a triangle pointing
;; inwards is by definition an output port, no matter whether it represents
;; incoming MIDI data received by a driver or MIDI data output by an
;; application (et vice versa). 
;; I maintain this terminology and as a result, driver-out-id accesses rather
;; :|Aout| or :|Bout| ids of the apple midi driver than :|Ain | or :|Bin | as
;; one might think.  

;;;
;;; always return a list, since some drivers offer more than one choice

(defun %find-ids (id-list type)
  (let ((r (funcall (ecase type
		      (:time #'first)
		      (:input #'third)           ; ill never understand this
		      (:output #'second))
		  id-list)))
    (if (listp r)
	r
      (list r))))

;;(defun %time-ids (id-list) (%find-ids id-list :time)) ; unused
(defun %in-ids (id-list) (%find-ids id-list :input))
(defun %out-ids (id-list) (%find-ids id-list :output))






;;; -------------------------------------------------------------------------
;;;
;;;
;;; Add and remove ports
;;;
;;; %mm-add-port takes a port structure and returns its refnum or the constant
;;; :memFullErr or :tooManyPortsErr if the port could not be allocated.
;;; %mm-remove-port expects the refnum of the port that is to be removed and 
;;; returns NIL. 

(defun %mm-add-port (port)
  (ccl:rlet ((clkinfo :MIDIClkInfo 
                      :synctype 0 
                      :curtime 0 
                      :format (port-format port))
             (portparams :MIDIPortParams 
                         :portid (port-id port)
                         :porttype (port-type port)
                         :timebase (if (eq (port-name port) :|Time|)
                                     0
                                     (let ((time (client-time-port %client%)))
                                       (or (port-refnum time)
                                           (error "Time port: no refnum"))))
                         :offsettime (port-offset port)
                         :readhook (ccl:%null-ptr)
                         :refcon 0
                         :initclock clkinfo
                         :name (string (port-name port)))
             (refnum :signed-integer))
    ;; Error handling here depends on the calling function: %mm-open 
    ;; signals an error, while %mm-make-aux-buffers continues w/o allocating
    ;; new ports.
    (let ((flg (#_midiaddport
		+client-id+ +buffer-size+ refnum portparams)))
      (case flg
        ((0 -255)                         ; #$noErr, #$MIDIvConnectMade
	 (setf (port-refnum port) (ccl:%get-signed-word refnum)))
	;; don't return numerical values for error codes, since it could be 
	;; misunderstood and taken as a refNum! 
        (-108 :memFullErr)
        (-252 :tooManyPortsErr)
        (t (warn "#_midiaddport failed with result code ~a." flg)
           (ccl:%get-signed-word refnum))))))

(defun %mm-remove-port (refnum)
  (if (numberp refnum)
    (#_midiremoveport refnum) 
    (error "can't remove a port with refNum ~s" refnum))
  ;; If the removed port was one of the main ports, clear its refNum.
  (let ((port (find refnum (client-static-ports %client%)
		    :key #'port-refnum)))
    (declare (dynamic-extent port))
    (when port
      (setf (port-refnum port) nil))))

;;;
;;; Allocate the static ports

(defun %mm-add-cm-ports ()
  (declare (special +mmCMSetUp+))
  ;; Add ports only if neccessary, so ask the MidiManager first what our ports
  ;; are -- but don't ask why they are still around.
  (let ((port-ostypes (%mm-get-midiidlist
		       (#_midigetports +client-id+))))
    (loop for port in (client-static-ports %client%)
          ;; NIL or an OSType, if the port is already present.
          for id = (find (port-id port) port-ostypes)
          with err = 0
          do 
          (when (not id)
	    (setf err (%mm-add-port port))
            ;; Ignore noErr and MIDIvConnectionMade result codes.
            (when (eq err :memFullErr)
              (error "MIDI: There is not enough memory to create a port!"))
	                (setf (port-refnum port) err))))
  (%dispatch :mmCMSetUp))

;;;
;;; Remove the static ports.  This also takes care not to remove ports that
;;; have already been removed to avoid bus errors.

(defun %mm-remove-cm-ports ()
  (declare (special +mmCMSignedIn+))
  (loop for port in (client-static-ports %client%)
        do (when (port-refnum port)
             (%mm-remove-port (port-refnum port))
             (setf (port-refnum port) nil)))
  (%dispatch :mmCMSignedIn))







;;; -------------------------------------------------------------------------
;;;
;;; Dynamic Port Handling
;;;
;;; %mm-make-dynamic-inport is the allocation routine. Invalidation takes
;;; place at interrupt level by MarkObsInPorts().  RemoveObsPorts() then does
;;; the actual deallocation work at non_interrupt time (see pmcl-midi.cc).

(defconstant +callback-period+ 
  ;; Time interval for the periodic MidiWakeUp callback. As long as we are
  ;; signed in, MarkObsInPorts() checks at these (millisecond) intervals
  ;; for empty buffers to mark as obsolete.
  5000)

(defconstant +buffer-idle-timeout+ 
  ;; Threshold (in ms) which determines after which time a buffer will be 
  ;; considered obsolete.
  (/ +callback-period+ 2))

(defvar %next-buffer-identifier% 
  ;; Simply counts up from 0. Only the low three bytes are needed.
  (the fixnum 0))

(defvar %dynamic-port-refnums%
  ;; Pointer to the DynamicPorts::fInvInPorts[0] array of kMaxInvInports ints.
  ;; kMaxInvInports is a constant defined in pmcl-midi.hh, but we don't have
  ;; to know its value since the array of valid port refnums is always
  ;; terminated by a 0 `fill-pointer' marker field.
  nil)

(defvar %dynamic-ports%
  ;; A-List that maps refnums of invisible ports to OSTypes.  This list is
  ;; maintained by %mm-make-dynamic-inport.
  nil)

(defvar %do-not-remove%
  ;; Pointer to the DynamicPorts::bOutputInProgress data member.
  nil)

(defvar %next-inport-id% 
  ;; Simply counts up from 0. Only the low three bytes are needed.
  (the fixnum 0))

(defun new-inport-name-and-id ()
  ;; Returns a namestring for a new invisible inport
  (let ((num (logand %next-inport-id% #xffffff)) name ostype)
    (setf name (concatenate 'string '"i " (princ-to-string num))
	  ostype (logior #x69000000 num)) ; #x69 = ASCII for #\i
    (incf %next-inport-id% 1)
    (values name ostype)))



;;;
;;; Add a new dynamic inport.  Returns the refnum of the newly allocated
;;; buffer (unused) or an error.

(defun %mm-make-dynamic-inport ()
  (let ((new (make-port :type (logior #$midiPortTypeInput 
				      $midiInvisiblePort))))
    (declare (dynamic-extent new))
    (multiple-value-bind (name ostype) (new-inport-name-and-id)
      (setf (port-name new) name
            (port-id new) ostype)
      ;; We only use refNums beyond this point.
      (let ((refnum (%mm-add-port new)))
	(when (or (eq refnum :memFullErr) (eq refnum :tooManyPortsErr))
	  ;; return :memFullErr no matter whether memory or MidiManager is full
	  (return-from %mm-make-dynamic-inport :memFullErr))
	;; Everything's ok, so register the new port
	(RegisterInvInPort refnum)
	;; Connect new inport to our invoutport
	(with-midi-checked 
	 (#_midiconnectdata 
	  +client-id+ +client-inv-outport-id+
	  +client-id+ ostype))
	;; Set the new inport's ReadHook to our invisibleReadHook
	(SetReadHook refnum)
	;; register the new refnum and ID with %dynamic-ports%
	(let ((entry (assoc refnum %dynamic-ports%)))
	  (if entry
	      (rplacd entry ostype)
	    (setf %dynamic-ports% (acons refnum ostype %dynamic-ports%))))
	refnum))))



;;; -------------------------------------------------------------------------
;;;
;;; %MM-Function Swapper
;;;
;;; The dispatch mechanism relies on the global semaphore %mm-gestalt%, which
;;; encodes security-relevant information about the current MidiManager
;;; environment and thus controls what may be done in midi at any given time.
;;;
;;; We have five states of Midi Manager "privileges" each of which allows
;;; only a specific subset of MidiManager calls to be made: 
;;;
;;;   :mmUnavailable, :mmUnknown, :mmAvailable, :mmCMSignedIn and :mmCMSetUp

(defvar %mm-gestalt% :mmUnknown)

(defun %mm-error (&rest rest)
  (declare (ignore rest))
  (ecase %mm-gestalt%
    (:mmUnavailable
     (error "MIDI: MidiManager is required to run Common Music with ~%~
             Midi syntax.~%~
             Make sure the MidiManager extension and a known Midi driver~%~
             are at their proper locations inside the System Folder and~%~
             reboot.~%~
             Currently known Midi drivers:~%~
             ~{~3T~a~%~}" (mapcar #'driver-filename %drivers%)))
    ((:mmUnknown :mmAvailable)
     (cerror "let Common Music open a midi port."
             "MIDI: There is no port open for midi.")
     (if (y-or-n-p "Do you want to specify a port?")
       (progn 
         (format t "~&Midi port to open (nil or a port name or number): ")
         (%midi-open
	  (loop for form = (read)
		while (and form
			   (not (find form %portrefs% :test #'member)))
		do (format t "~s is not a known port specifcation. ~
                              Midi port to open: " form)
		finally (return form)))) ; may be nil
       (%midi-open nil)))
    (:mmCMSignedIn
     (error "MIDI: oops! Common Music has not yet any ports allocated.~%~
             Please try closing Midi and opening it again."))
    (:mmCMSetUp
     (error "MIDI: oops! <strange error: MidiManager Gestalt value: ~a>" 
	    %mm-gestalt%))))
     

;;;
;;; Mapping of %midi-functions onto %mm-functions

(defconstant %midi/mm-siblings%
  '(;; %midi-open is the entry function and thus always available
    (%midi-close 		%mm-close) 
    (%midi-start-time 		%mm-start-time) 
    (%midi-stop-time 		%mm-stop-time) 
    (%midi-get-time 		%mm-get-time) 
    (%midi-set-time 		%mm-set-time) 
    (%midi-read-messages 	%mm-read-messages) 
    (%midi-write-message 	%mm-write-message) 
    (%midi-all-notes-off 	%mm-all-notes-off) 
    (%midi-flush-transmit 	%mm-flush-transmit) 
    (%midi-flush-receive 	%mm-flush-receive) 
    (%midi-hush 		%mm-hush)))

(defun %level-3-siblings () (nthcdr 1 %midi/mm-siblings%))
(defun %level-2-siblings () (list (first %midi/mm-siblings%)))


;;;
;;; Initialization

(eval-when (:load-toplevel :execute)
  (map nil #'(lambda (x) (setf (symbol-function (first x)) #'%mm-error))
       %midi/mm-siblings%))


(defun %dispatch (new-gestalt)
  (ecase %mm-gestalt%
    (:mmAvailable 
     ;; new-gestalt is always :mmCMSignedIn (can't go back to :mmUnknown)
     (map nil #'(lambda (x)
		  (setf (symbol-function (first x))
			(symbol-function (second x))))
          (%level-2-siblings)))
    (:mmCMSignedIn
     (ecase new-gestalt
       ;; new-gestalt is :mmAvailable if we signed out or :mmCMSetUp if all
       ;; is up and running now 
       (:mmAvailable
	(map nil #'(lambda (x) 
		     (setf (symbol-function (first x)) #'%mm-error))
	     (%level-2-siblings)))
       (:mmCMSetUp
	(map nil #'(lambda (x)
		     (setf (symbol-function (first x)) 
			   (symbol-function (second x))))
	     (%level-3-siblings)))))
    (:mmCMSetUp
     ;; new-gestalt is always :mmCMSignedIn (can't go further)
     (map nil #'(lambda (x)
		  (setf (symbol-function (first x)) 
			(symbol-function '%mm-error)))
          (%level-3-siblings))))
  (setf %mm-gestalt% new-gestalt))








;;; -------------------------------------------------------------------------
;;;
;;; Signing In and Out
;;;



(defvar %iconhandle% nil)

#-cm
(defvar *cm-resources* nil)

(eval-when (:load-toplevel :execute)
  (let ((resfile (merge-pathnames ".rsrc" ccl:*loading-file-source-file*)))
    (ccl:with-open-resource-file (ref resfile)
      (ccl:using-resource-file ref 
        (loop for type in '(:|ICN#| :|icl4| :|icl8|)
              with id = #x4100      ; hardcoded
              for res = (list (let ((r (ccl:require-trap _getresource 
                                                         type id)))
                                (ccl:require-trap _loadresource r)
                                (ccl:require-trap _detachresource r)
                                r)
                              type id)
              when res
              do (pushnew res *cm-resources*)
              else do (warn "Resource ~a ID ~d not found in ~a"
                            (symbol-name type) id resfile))))))



;;;
;;; %mm-sign-in checks for the presence of some client with the same id as 
;;; itself. If such a client is found and its name is _not_ equal to 
;;; (cm-version), a continuable error is signaled, leaving the choice
;;; to the user. Else it assumes that for some reason Common Music couldn't
;;; sign out properly and simply returns nil. If it did actually sign in, 
;;; however, the return value is t.

(defun %mm-sign-in ()
  (let ((refcon 0))
    (ccl:with-pstrs ((name +client-name+))
      ;; Try to sign in normally.
      (case (#_midisignin +client-id+ refcon 
             (or %iconhandle%
                 (setf %iconhandle% (#_getresource :|ICN#| #x4100))) name)
        (0
	 ;; #$noErr, Return t.
         (%dispatch :mmCMSignedIn) 
         t)
        (-108
	 ;; #$memFullErr
         (error "MIDI: not enough memory to sign into the MidiManager"))
        (-260
	 ;; #$MIDIDupIDErr
         ;; There is already another client with the same id.  To get its
	 ;; name, I have to sign in temporarily with another id.
         (let (id err)
           (loop for randid = (nth-value 1 (new-inport-name-and-id))
                 for flg = (#_midisignin randid refcon %iconhandle% name)
                 thereis (when (/= flg #$MIDIDupIDErr)  ; if it's not already
                           (setf id randid)     	;   there, exit loop
                           (setf err flg)
                           (%dispatch :mmCMSignedIn)))
           (case err
             (0
              ;; #$noErr, success. Now get the conflicting client's
              ;; name and see whether it's Common Music.
              (if (ccl:%stack-block ((str255 256))
                    (#_midigetclientname +client-id+ str255)
                    (string-equal (ccl:%get-string name) 
                                  (ccl:%get-string str255)))
                ;; It's us, so don't do anything.
                nil
                ;; Else leave the choice to the user...
                (cerror "Assume that the client is ~a~@
                         (Which it does not seem to be)."
                 "MIDI: duplicate client ID - apparently not Common Music.~@
                  If there is another application with the same ID, you will~@
                  have to quit it before Common Music can sign in.~@
                  However, if you know that Common Music is just still ~@
                  signed in for some reason, continue at your own risk."
                  name))
              ;; Sign out again and return nil.
              (#_midisignout id)
              nil)
             ;; Else give up.
             (t (error "MIDI: While trying to recover from a ~
                        DuplicateClientID error:~@
                        Failed to sign in using a different ID with ~a" err)
                ;; Sign out again and return nil.
                (#_midisignout id)
                nil))))
        (t
	 ;; There shouldn't be any other result codes, but make sure anyway.
	 (error "MIDI: failed to sign in with an unknown error."))))))



(defun %mm-sign-out ()
  (#_midisignout +client-id+)
  (%dispatch :mmAvailable)
  (#_disposehandle %iconhandle%)
  t)


;;; -------------------------------------------------------------------------
;;;
;;; MidiManager Connections
;;;
;;; This functionality could be extended in the future to make MidiManager 
;;; connections scriptable.


(defun %mm-connect (driver-ports)    ; NIL, 0 (:A), 1 (:B), 2 (:A+B) or 3 (:B+A)
  (when driver-ports
    (let* ((driver (find-driver-by-id %driver%))
           (1st-ids (ecase driver-ports
		      ((0 :A 2 :A+B) (driver-a-ids driver))
                      ((3 :B+A) (driver-b-ids driver))
                      ((1 :B) nil)))
           (2nd-ids (ecase driver-ports
                      ((1 :B 2 :A+B) (driver-b-ids driver))
                      ((3 :B+A) (driver-a-ids driver))
                      ((0 :A) nil))) 
           outids-1 outids-2 inids-1 inids-2)
      ;; clear the second set if it's the same.
      (when (eq 1st-ids 2nd-ids)
        (setf 2nd-ids nil))
      (setf outids-1 (and 1st-ids (%out-ids 1st-ids))
	    outids-2 (and 2nd-ids (%out-ids 2nd-ids))
	    inids-1 (and 1st-ids (%in-ids 1st-ids))
	    inids-2 (and 2nd-ids (%in-ids 2nd-ids)))
      (when 1st-ids
        (when outids-1
          (dolist (oi outids-1)
	    (with-midi-checked
	      (#_midiconnectdata %driver% oi
               +client-id+ (port-id 
                            (client-inport-A %client%))))))
        (when inids-1
          (dolist (ii inids-1)
	    (with-midi-checked
	      (#_midiconnectdata +client-id+ (port-id 
                                              (client-outport-A %client%))
               %driver% ii))))
        (when 2nd-ids
          (when outids-2
            (dolist (oi outids-2)
	      (with-midi-checked
	        (#_midiconnectdata %driver% oi
                 +client-id+ (port-id 
                              (client-inport-B %client%))))))
          (when inids-2
            (dolist (ii inids-2)
	      (with-midi-checked
	        (#_midiconnectdata +client-id+ (port-id 
                                                (client-outport-B %client%))
                 %driver% ii)))))))
    t))

#|
;;;
;;; Dummy - disconnecting is achieved automatically by signing out.

(defun %mm-disconnect (&optional driver-ports)
  (declare (ignore driver-ports))
  t)
|#







;;; -------------------------------------------------------------------------
;;;
;;; Time 
;;;


(defun %mm-start-time ()
  (#_midistarttime (port-refnum (client-time-port %client%)))
  (values +midi-success+))

(defun %mm-stop-time ()
  (#_midistoptime (port-refnum (client-time-port %client%)))
  (RemoveObsPorts)
  (values +midi-success+))

(defun %mm-get-time ()
  (RemoveObsPorts)
  (values (#_midigetcurtime (port-refnum (client-time-port %client%)))))

(defun %mm-set-time (time)
  (#_midisetcurtime (port-refnum (client-time-port %client%)) time)
  (RemoveObsPorts)
  (values +midi-success+))





;;; -------------------------------------------------------------------------
;;;
;;; Midi I/O 
;;;

(defun %mm-read-messages ()
  (MidiReadEvent)
  (RemoveObsPorts)
  (values (the fixnum 0)))



;;;
;;; msg is an already set up stack block of size 16 containing the message.
;;; Return the refnum of the invisible inport the message has been written
;;; to. 

(defun %mm-write-message-aux (msg)
  ;; no-more-alloc = nil, i = -1
  ;; loop {
  ;;   refnum = %dynamic-port-refnums%[++i]
  ;;   if (refnum == kNoPort)
  ;;     if (no-more-alloc) reset i and break
  ;;     else
  ;;       allocate new port
  ;;       if (failure)
  ;;         no-more-alloc = t
  ;;         if (i == 0) error "no port exists and can't allocate new port"
  ;;         else reset i and break
  ;;       else
  ;;         refnum = %dynamic-port-refnums%[0], reset i
  ;;         WRITE
  ;;         if (failure) error "can't write to new buffer!"
  ;;         else return refnum
  ;;   else 
  ;;     disconnect old refnum, connect refnum to +inv-outport+
  ;;     WRITE
  ;;     if (no failure)
  ;;       make this refnum the current port
  ;;       return refnum
  ;; }
  (loop for offs = 0 then (+ offs 2)	; advance in shorts
	with cur-refnum = (ccl:%get-word %dynamic-port-refnums% 0)
        for refnum = cur-refnum then (ccl:%get-word 
                                      %dynamic-port-refnums% offs)
	with no-more-alloc = nil
	with err
	do (if (= refnum 0)
	       (if no-more-alloc 
                 ;; start over.  do NOT update cur-refnum.
                 ;; this effectively blocks, so call ccl:event-dispatch
                 (progn
                   (setf offs -2)
                   (ccl:event-dispatch t))
		 (progn
		   (setf err (%mm-make-dynamic-inport))
		   (if (eq err :memFullErr)
		       (progn
			 (setf no-more-alloc t)
			 (if (= offs 0)
			     (error "(shouldn't) no port exists and can't ~
                                     allocate new ports!")
                             ;; start over.  do NOT update cur-refnum.
			   (setf offs -2)))
		     ;; else write to new port
		     (progn
		       ;; the newly allocated port is at position 0 and
		       ;; already connected.  We could also assume to be
		       ;; able to write to it without checking for errors
		       ;; since it's empty, but this seems safer...
                       ;; disconnect cur-refnum
                       (when (/= cur-refnum 0)
 	                 (with-midi-checked 
		           (#_midiunconnectdata 
		            +client-id+ +client-inv-outport-id+
		            +client-id+ (cdr (assoc cur-refnum 
                                                    %dynamic-ports%)))))
		       (setf refnum (ccl:%get-word %dynamic-port-refnums% 0)
			     cur-refnum refnum
                             offs -2	; we are back to square one
			     err (#_midiwritepacket +inv-outport-refnum+ msg))
		       (unless (= err #$noErr)
			 (error "(shouldn't) can't write to new buffer ~
                                     ~a! (Error Code ~a)" refnum err))
		       (return-from %mm-write-message-aux refnum)))))
	     ;; else connect it to +client-inv-outport-id+ and try to
	     ;; write to it
	     (progn
               (unless (= refnum cur-refnum)
	         (with-midi-checked 
		   (#_midiunconnectdata 
		    +client-id+ +client-inv-outport-id+
		    +client-id+ (cdr (assoc cur-refnum %dynamic-ports%))))
	         (with-midi-checked 
		   (#_midiconnectdata 
		    +client-id+ +client-inv-outport-id+
		    +client-id+ (cdr (assoc refnum %dynamic-ports%))))
                 (setf cur-refnum refnum))
	       (when (= (#_midiwritepacket +inv-outport-refnum+ msg) #$noErr)
                 ;; make this buffer current
                 (when (> offs 0)
                   (RotateInvInPorts (floor offs 2)))
                 (return-from %mm-write-message-aux refnum))))))


(defun %mm-write-message (message time)
  (declare (optimize (speed 3)(safety 0))
           (fixnum message time)
	   (inline %mm-write-message-aux))
  (let ((size (ash (logand message #x03000000) -24)) refnum)
    (declare (fixnum byte size))
    (when (or (< size 1) (> size 3)) (error "bad message size: ~A" size))
    ;; Allocate a stack block in the mac heap and put the midi packet in it.
    (ccl:%stack-block ((buf 16))
      (ccl:%put-byte buf 0 0)           ; flags
      (ccl:%put-byte buf (+ size 6) 1)  ; size
      ;; offset 5 (msb of message) gets overwritten below by lsb of time (msb
      ;; of message is unused) 
      (ccl:%put-long buf message 5)
      (ccl:%put-long buf time 2)

      (ccl:%put-word %do-not-remove% #x0001) ; block MarkObsInPorts()
      (setf refnum (%mm-write-message-aux buf))
      ;; Store this message's time in the current invisible inport
      (with-midi-checked (#_midisetrefcon refnum time))
        
      (ccl:%put-word %do-not-remove% #x0000) ; unblock MarkObsInPorts()
      (RemoveObsPorts)
      +midi-success+)))







;;; -------------------------------------------------------------------------
;;;
;;; Flushing 
;;;

(defun %mm-flush-transmit ()
  (FlushQueues)
  (values +midi-success+))

(defun %mm-flush-receive ()
  (#_midiflush (port-refnum (client-inport-a %client%)))
  (#_midiflush (port-refnum (client-inport-b %client%)))
  (values +midi-success+))

(defun %mm-hush ()
  ;; Flush all pending midi input and output events.
  (%mm-flush-transmit)
  (%mm-flush-receive)
  (values +midi-success+))



;;; 
;;; Output note-off messages for all keys and all channels.
;;; A stack block is set up as a MidiPacket with the fields:
;;;
;;;                  Byte:  0     1      2-5   6-8 (= 3 lsb of data)
;;;                 Field:  flags length time  data
;;;        initialized to:  0009  0000   0000  8xxx  7f.. .... .... ....
;;;
;;; Only the word at offset 6 (the 2 MSB of data) is chnged.

(defun %mm-all-notes-off (time)
  (declare (optimize (speed 3)(safety 0))
           (ignore time))
  (let ((outa-refnum (port-refnum (client-outport-a %client%)))
	(outb-refnum (port-refnum (client-outport-b %client%))))
    (ccl:%put-word %do-not-remove% #x0001)
    (ccl:%stack-block ((buf 16))
      (ccl:%put-word buf #x0009 0)
      (ccl:%put-long buf #x00800000 5)
      (ccl:%put-long buf #x00000000 2)
      (loop with time = (%mm-get-time)
            for channel from #x80 to #x81
	    do (progn
		 (ccl:%put-byte buf channel 6)
		 (loop for note from #x00 to #x7f 
                       do (progn 
                            (incf time)
                            (loop until (>= (%mm-get-time) time))
                            (ccl:%put-byte buf note 7)
		            (#_midiwritepacket outa-refnum buf)
		            (#_midiwritepacket outb-refnum buf)))))
      (ccl:%put-word %do-not-remove% #x0000)
      (RemoveObsPorts)
      (values +midi-success+))))



;;; -------------------------------------------------------------------------
;;;
;;; Opening and closing midi
;;;
;;; %midi-open is the midi entry function. It performs its own checks on 
;;; whether the MidiManager has been installed, etc. Therefore, this function
;;; is accessible all the time and will never swapped.

;;;
;;; Foreign function environment initialization 

(defun midi-c-setup () 
  (unless %lisp-midi-read-hook%
    (setf %lisp-midi-read-hook% 
          (ccl:pref (ccl:pref %midi-read-hook% :RoutineDescriptor.RoutineRecords)
                    :RoutineRecord.ProcDescriptor)))
  (setf %dynamic-port-refnums%
        (MidiCSetUp (port-refnum (client-time-port %client%))
                    (port-refnum (client-inport-A %client%))
                    (port-refnum (client-inport-B %client%))
                    (port-refnum (client-outport-A %client%))
                    (port-refnum (client-outport-B %client%))
                    (port-refnum (client-inv-outport %client%))
                    %lisp-midi-read-hook% ;;%midi-read-hook%
                    +callback-period+
                    +buffer-idle-timeout+ 
                    *midi-flush-timeout*))
  (if (ccl:%null-ptr-p %dynamic-port-refnums%)
      (progn
	(warn "(shouldn't) midi-c-setup failed.  Closing Midi.")
	nil)
    (progn
      (setf %do-not-remove% (GetbOutputInProgress))
      t)))


(defun check-for-mm-availability ()
  ;; Set it this one time directly - don't call %dispatch at this level.
  (setf %mm-gestalt% (if (= (#_snddispversion #$midiToolNum) 0) 
                       :mmUnavailable 
                       :mmAvailable))
  ;; Don't return to %mm-open unless the MidiManager is available.
  (when (eq %mm-gestalt% :mmUnavailable) (%mm-error))
  t)
  
(defvar %mm-close-stack%
  ;; list of functions to be executed by %mm-close.  This gets filled by
  ;; %midi-open 
  nil)

(defun %mm-close ()
  (map nil #'funcall %mm-close-stack%)
  (setf %mm-close-stack% nil)
  #+cm (setf *midi-open* nil) ; communicate status back to CM
  (values +midi-success+))


;;;
;;; %mm-get-driver-and-ports
;;; ------------------------
;;; Look for the preferred driver and driver ports (depending on
;;; *midi-preferences*).  We set the globals %driver% and %driver-ports% here. 
;;; I assume the driver is the most important point and then try to further
;;; satisfy the port requests.  This may or may not be the right way to do it.

(defun %mm-get-driver-and-ports (&optional ports)
  (let ((port-prefs (if ports
			(cons ports (second *midi-preferences*))
		      (second *midi-preferences*)))
	(mm-drivers (%mm-get-midiidlist (#_midigetclients)))
	mm-driver-ports)
    ;; translate port-prefs to numbers
    (mapl #'(lambda (c)
	      (setf (car c) (%translate-portref (car c) :number)))
	  port-prefs)
    ;; get the most preferred driver
    (loop for key in (first *midi-preferences*)
	  for driver = (find-driver-by-name key)
	  for mm-driver = (and driver (first (member (driver-id driver)
						     mm-drivers)))
	  do (when mm-driver
	       ;; if a known driver is available, check its can-connect field
	       (case (driver-can-connect driver)
		 (nil (warn "Selecting a driver that does not support ~
                               automatic connecting. You will have to make~
                               your connections manually using PatchBay")
		      (setf %driver-ports% nil ; don't connect
			    %driver% mm-driver))
		 ;; else check the ports we should connect to
		 (t (setf mm-driver-ports
			  (%mm-get-midiidlist (#_midigetports mm-driver)))
		    (flet ((port-exists (id)
					(if id
					    (find id mm-driver-ports)
					  t)))
		      (let* ((a-ids (driver-a-ids driver))
			     (b-ids (driver-b-ids driver))
			     a? b?)
			;; resolve the OMS port id's.  The a-or-b question is
			;; unimportant for OMS
			(if (eq mm-driver :|OmMd|)
			    (progn
			      (mapl #'(lambda (cs)
					(when (functionp (car cs))
					  (setf (car cs)
						(funcall (car cs)))))
				    a-ids) ; b-ids is structure-shared
			    (setf %driver-ports% 0)) ; anything simple
			  (progn
			    (setf a? (every #'port-exists a-ids)
				  b? (every #'port-exists b-ids))
			    ;; loop trying to satisfy the *midi-preferences*
			    (loop for pref in port-prefs
				  do (when (ecase pref
					     (0 a?) (1 b?) (2 (and a? b?)))
				       (setf %driver-ports% 
                                             (if (and (= pref 2)
                                                      (eq a-ids b-ids))
                                               0        ; anything simple
                                               pref))
				       (return)))
			    ;; else connect to whatever ports are open
			    (setf %driver-ports% 
                                  (if (and a? b?) (or (and (eq a-ids b-ids)
                                                           0)   ; sth simple
                                                      2) 
                                      (if b? 1
                                          (if a? 0
                                              nil))))))
			(setf %driver% mm-driver)))))
	       (return)))
    (unless %driver%
      (warn "MIDI: no known midi driver with open ports available.~@
             Check the cm:*midi-preferences* parameter and/or PatchBay to~@
             see whether your driver is included and has any ports enabled.~@
             ~%Type y(es) if you could fix the problem, or n(o) to cancel.~%")
      (if (y-or-n-p "~%Retry opening/connecting to a midi driver?")
	  (%mm-get-driver-and-ports ports)
        (throw :midi-open-exit :midi-open-cancelled)))))


;;; Arguments to %midi-open:  
;;;             -1  ports unspecified (do what seems reasonable)
;;;              0  open port A
;;;              1  open port B
;;;              2  open port A+B

(defun %mm-midi-open-aux (int)
  (catch :midi-open-exit
    (setf %mm-close-stack% 
          `(,#'(lambda ()
                 (setf %midi-open% nil)
                 (%update-midi-menu)   ; set below
                 (values +midi-success+)))) 
    (%mm-sign-in)
    (push #'%mm-sign-out %mm-close-stack%)
    (%mm-add-cm-ports)
    (push #'(lambda () 
              (%mm-remove-cm-ports)
              ;; invalidate the pointers held in 
              (setf %dynamic-ports% nil)
              (when  %dynamic-port-refnums%
                (setf %dynamic-port-refnums% nil)))
	  %mm-close-stack%)
    ;; determine and set %driver% and %driver-ports%.
    (%mm-get-driver-and-ports int)
    (if %driver-ports%
      (progn
        (push #'(lambda ()
	          (setf %driver-ports% nil %driver% nil))
              %mm-close-stack%)
        (%mm-connect %driver-ports%)
        (%midi-start-time)
        ;; make sure %midi-open% holds the serial port variable
        (setf %midi-open% (%translate-portref %driver-ports% :long))
        (%update-midi-menu)
        ;;(push #'%update-midi-menu %mm-close-stack%)
        ;; Initialize the c side; but close everything again, if it failed.
        (unless (midi-c-setup) 
          (throw :midi-open-exit :midi-open-cancelled))
        (values +midi-success+))
      (progn
        (warn "The preferred driver: ~%   ~s~%does not have any ports open.~%~
               Enable a port in PatchBay and re-open MIDI."
              (driver-filename (find-driver-by-id %driver%)))
        (%mm-close)
        (values -1)))))


(defun %midi-open (int)
  ;; I need nil instead of -1 for %mm-get-driver-and-ports 
  (when (eql int -1) (setf int nil))	
  ;; test for MidiManager's presence the first time only
  (case %mm-gestalt%
    (:mmUnavailable (%mm-error))
    (:mmUnknown (check-for-mm-availability)))
  (if (eq (%mm-midi-open-aux int) :midi-open-cancelled)
    (progn
      (%mm-close)
      -1)				; canceled
    (progn
      (setf +inv-outport-refnum+ (port-refnum +inv-outport+))
      #+cm (setf *midi-open* %midi-open%) ; communicate status back to CM
      +midi-success+)))




;;;
;;; -*- EOF -*-
